import os

from .constant import (
    CASTOR_LOOKER_PAGE_SIZE,
    CASTOR_LOOKER_TIMEOUT_SECOND,
    DEFAULT_LOOKER_PAGE_SIZE,
    DEFAULT_LOOKER_TIMEOUT_SECOND,
)


def _int_env(key_env_name: str, default_value: int) -> int:
    value = os.environ.get(key_env_name, default_value)
    try:
        return int(value)
    except ValueError:
        return default_value


def page_size() -> int:
    """
    Either returns the page size parameter from env or from default value
    """
    return _int_env(
        key_env_name=CASTOR_LOOKER_PAGE_SIZE,
        default_value=DEFAULT_LOOKER_PAGE_SIZE,
    )


def timeout_second() -> int:
    """
    Either returns the timeout second parameter from env or from default value
    """
    return _int_env(
        key_env_name=CASTOR_LOOKER_TIMEOUT_SECOND,
        default_value=DEFAULT_LOOKER_TIMEOUT_SECOND,
    )



from abc import abstractmethod
from enum import Enum
from itertools import chain
from typing import (
    Callable,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    TypeVar,
)
from uuid import UUID

_DEFAULT_PER_PAGE = 100
_DEFAULT_MIN_UUID = UUID("00000000-0000-0000-0000-000000000000")

T = TypeVar("T")


class PagerLogger:
    def on_page(self, page: int, count: int):
        pass

    def on_success(self, page: int, total: int):
        pass


class PagerOnIdLogger:
    def on_iteration(self, max_id: UUID, count: int):
        pass

    def on_success(self, total: int):
        pass


class PagerStopStrategy(Enum):
    """Strategy for stopping the pager"""

    EMPTY_PAGE = "EMPTY_PAGE"
    LESS_RESULTS_THAN_ASKED = "LESS_RESULTS_THAN_ASKED"


class AbstractPager(Generic[T]):
    def all(self, per_page: int = _DEFAULT_PER_PAGE) -> List[T]:
        """Returns all data provided by the callback as a list"""
        return list(chain.from_iterable(self.iterator(per_page=per_page)))

    @abstractmethod
    def iterator(self, per_page: int) -> Iterator[Sequence[T]]:
        pass


class Pager(AbstractPager):
    def __init__(
        self,
        callback: Callable[[int, int], Sequence[T]],
        *,
        logger: Optional[PagerLogger] = None,
        start_page: int = 1,
        stop_strategy: PagerStopStrategy = PagerStopStrategy.EMPTY_PAGE,
    ):
        self._callback = callback
        self._logger = logger or PagerLogger()
        self._start_page = start_page
        self._stop_strategy = stop_strategy

    def iterator(
        self, per_page: int = _DEFAULT_PER_PAGE
    ) -> Iterator[Sequence[T]]:
        """Yields data provided by the callback as a list page by page"""
        page = self._start_page
        total_results = 0

        stop_on_empty_page = self._stop_strategy == PagerStopStrategy.EMPTY_PAGE

        while True:
            results = self._callback(page, per_page)
            nb_results = len(results)
            total_results += nb_results
            is_empty = nb_results == 0
            is_partial_page = nb_results < per_page
            should_stop = is_empty if stop_on_empty_page else is_partial_page

            self._logger.on_page(page, nb_results)
            if results:
                yield results

            if should_stop:
                break

            page += 1

        self._logger.on_success(page, total_results)


class PagerOnId(AbstractPager):
    def __init__(
        self,
        callback: Callable[[UUID, int], Sequence[T]],
        *,
        logger: Optional[PagerOnIdLogger] = None,
        stop_strategy: PagerStopStrategy = PagerStopStrategy.EMPTY_PAGE,
    ):
        self._callback = callback
        self._logger = logger or PagerOnIdLogger()
        self._stop_strategy = stop_strategy

    def iterator(
        self, per_page: int = _DEFAULT_PER_PAGE
    ) -> Iterator[Sequence[T]]:
        """Yields data provided by the callback as a list using the greatest UUID as a reference point"""
        greater_than_id = _DEFAULT_MIN_UUID
        stop_on_empty_page = self._stop_strategy == PagerStopStrategy.EMPTY_PAGE

        total_results = 0

        while True:
            results = self._callback(greater_than_id, per_page)
            nb_results = len(results)
            total_results += nb_results
            is_empty = nb_results == 0
            is_partial_page = nb_results < per_page
            should_stop = is_empty if stop_on_empty_page else is_partial_page
            if results:
                greater_than_id = self._max_id(results)
                self._logger.on_iteration(greater_than_id, nb_results)
                yield results

            if should_stop:
                break
        self._logger.on_success(total_results)

    @staticmethod
    def _max_id(items: Sequence) -> UUID:
        return max(item["id"] for item in items)



import os.path
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client import ApiClient
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "../assets/rest_api/")

AUTH_XML = os.path.join(TEST_ASSET_DIR, "auth.xml")


class AuthTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient()
        self.baseurl = self._client._server.auth.baseurl

    def test_auth(self):
        with open(AUTH_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl + "/signin", text=response_xml)
            self._client.login()

        self.assertEqual(
            "eIX6mvFsqyansa4KqEI1UwOpS8ggRs2l", self._client._server.auth_token
        )
        self.assertEqual(
            "6b7179ba-b82b-4f0f-91ed-812074ac5da6", self._client._server.site_id
        )
        self.assertEqual(
            "1a96d216-e9b8-497b-a82a-0b899a965e01", self._client._server.user_id
        )



from .decryption import decrypt


def test_decrypt():

    key = "8iY5gyHxPH0YYyBgOd2AvwT1pcHl3EGtvN5jAi9JwoA"
    enc_message = "sIqpGK6UgzKOkThtlvBGqkb046EtB+HxcBsO3nDiKZAJcszfqqxTgSyH+SXAALznfuMSnZjdX9yzpGe77+ByYuCVlXHkMilkUe6tkFsFkBPW5CPirp0kqLdyp1yHXrv3NmXCtGZcef2fC0v89huRMSgFcm8M6Zf3JjSDEludLUo="
    dec_message = '{"db":"zip:/app/metabase.jar!/sample-dataset.db;USER=GUEST;PASSWORD=guest"}'

    dec = decrypt(enc_message, key)
    assert dec == dec_message

    key = "TkZFBIEqRM/C+qB20+lDv+zISuAXFVvBzsr5NrD6rgo="
    dec = decrypt(enc_message, key)
    assert dec != dec_message



from argparse import ArgumentParser

from castor_extractor.visualization import looker  # type: ignore


def main():
    parser = ArgumentParser()
    parser.add_argument("-b", "--base-url", help="Looker base url")
    parser.add_argument("-u", "--username", help="Looker client id")
    parser.add_argument("-p", "--password", help="Looker client secret")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-t", "--timeout", type=int, help="Timeout in seconds")
    parser.add_argument(
        "--all-looks",
        help="Use all_looks endpoint instead of the paginated search_looks",
        action="store_true",
    )
    parser.add_argument(
        "--safe-mode", "-s", help="Looker safe mode", action="store_true"
    )

    args = parser.parse_args()

    looker.extract_all(
        base_url=args.base_url,
        client_id=args.username,
        client_secret=args.password,
        output_directory=args.output,
        timeout=args.timeout,
        all_looks=args.all_looks,
        safe_mode=args.safe_mode,
    )



from typing import Set

from ..column import ColumnChecker
from ..constants import COLUMN_TYPES, TABLE_TYPES
from ..enums import DataType
from ..file import FileTemplate


class GenericWarehouseFileTemplate:
    """
    Provides templates to validate generic warehouse files:
    - catalog: databases, schemas, tables, columns
    - users
    - queries
    - views DDL
    """

    @staticmethod
    def database() -> FileTemplate:
        return {
            "id": ColumnChecker(is_unique=True),
            "database_name": ColumnChecker(),
        }

    @staticmethod
    def schema(database_ids: Set[str]) -> FileTemplate:
        return {
            "id": ColumnChecker(is_unique=True),
            "database_id": ColumnChecker(foreign=database_ids),
            "schema_name": ColumnChecker(),
            "description": ColumnChecker(is_mandatory=False),
            "tags": ColumnChecker(is_mandatory=False),
        }

    @staticmethod
    def table(schema_ids: Set[str]) -> FileTemplate:
        return {
            "id": ColumnChecker(is_unique=True),
            "schema_id": ColumnChecker(foreign=schema_ids),
            "table_name": ColumnChecker(),
            "description": ColumnChecker(is_mandatory=False),
            "tags": ColumnChecker(is_mandatory=False),
            "type": ColumnChecker(enum_values=TABLE_TYPES),
        }

    @staticmethod
    def column(table_ids: Set[str]) -> FileTemplate:
        return {
            "id": ColumnChecker(is_unique=True),
            "table_id": ColumnChecker(foreign=table_ids),
            "column_name": ColumnChecker(),
            "description": ColumnChecker(is_mandatory=False),
            "data_type": ColumnChecker(enum_values=COLUMN_TYPES),
            "ordinal_position": ColumnChecker(
                is_mandatory=False, data_type=DataType.INTEGER
            ),
        }

    @staticmethod
    def query(database_ids: Set[str], user_ids: Set[str]) -> FileTemplate:
        return {
            "database_id": ColumnChecker(foreign=database_ids),
            "query_text": ColumnChecker(),
            "user_id": ColumnChecker(foreign=user_ids),
            "start_time": ColumnChecker(data_type=DataType.DATETIME),
            "end_time": ColumnChecker(
                data_type=DataType.DATETIME, is_mandatory=False
            ),
        }

    @staticmethod
    def view_ddl() -> FileTemplate:
        return {
            "database_name": ColumnChecker(is_mandatory=True),
            "schema_name": ColumnChecker(is_mandatory=True),
            "view_name": ColumnChecker(is_mandatory=True),
            "definition": ColumnChecker(is_mandatory=True),
        }

    @staticmethod
    def user() -> FileTemplate:
        return {
            "id": ColumnChecker(is_unique=True),
            "email": ColumnChecker(is_mandatory=False),
            "first_name": ColumnChecker(is_mandatory=False),
            "last_name": ColumnChecker(is_mandatory=False),
        }



import json
import logging
import os

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError

from .....utils import SerializedAsset, from_env
from ...assets import EXPORTED_FIELDS, MetabaseAsset
from ...errors import EncryptionSecretKeyRequired, MetabaseLoginError
from ..decryption import decrypt
from ..shared import DETAILS_KEY, get_dbname_from_details
from .credentials import CredentialsDb, CredentialsDbKey, get_value

logger = logging.getLogger(__name__)

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

PG_URL = "postgresql://{username}:{password}@{host}:{port}/{database}"
SQL_FILE_PATH = "queries/{name}.sql"

ENCRYPTION_SECRET_KEY = "CASTOR_METABASE_ENCRYPTION_SECRET_KEY"


class DbClient:
    """
    Connect to Metabase Database and fetch main assets.
    """

    def __init__(
        self,
        **kwargs,
    ):
        self._credentials = CredentialsDb(
            host=get_value(CredentialsDbKey.HOST, kwargs),
            port=get_value(CredentialsDbKey.PORT, kwargs),
            database=get_value(CredentialsDbKey.DATABASE, kwargs),
            schema=get_value(CredentialsDbKey.SCHEMA, kwargs),
            username=get_value(CredentialsDbKey.USERNAME, kwargs),
            password=get_value(CredentialsDbKey.PASSWORD, kwargs),
        )
        self.encryption_secret_key = kwargs.get(
            "encryption_secret_key"
        ) or from_env(ENCRYPTION_SECRET_KEY, allow_missing=True)
        self._engine = self._login()

    def _login(self) -> Engine:
        url = PG_URL.format(**self._credentials.to_dict(hide=False))
        try:
            engine = create_engine(url)
            engine.connect()
            return engine
        except OperationalError as err:
            raise MetabaseLoginError(
                credentials_info=self._credentials.to_dict(hide=True),
                error_details=err.args,
            )

    def _load_query(self, name: str) -> str:
        """load SQL text from file"""
        filename = SQL_FILE_PATH.format(name=name)
        path = os.path.join(CURRENT_DIR, filename)
        with open(path, "r") as f:
            content = f.read()
            return content.format(schema=self._credentials.schema)

    def _database_specifics(
        self, databases: SerializedAsset
    ) -> SerializedAsset:
        for db in databases:
            assert DETAILS_KEY in db  # this field is expected in database table

            try:
                details = json.loads(db[DETAILS_KEY])
            except json.decoder.JSONDecodeError as err:
                if not self.encryption_secret_key:
                    raise EncryptionSecretKeyRequired(
                        credentials_info=self._credentials.to_dict(hide=True),
                        error_details=err.args,
                    )
                decrypted = decrypt(db[DETAILS_KEY], self.encryption_secret_key)
                details = json.loads(decrypted)

            db["dbname"] = get_dbname_from_details(details)

        return databases

    @staticmethod
    def name() -> str:
        """return the name of the client"""
        return "Metabase/DB"

    def base_url(self) -> str:
        """Fetches the `base_url` of the Metabase instance"""
        query = self._load_query(name="base_url")
        result = self._engine.execute(query)
        return result.fetchone()["value"]

    def fetch(self, asset: MetabaseAsset) -> SerializedAsset:
        """fetches the given asset"""
        query = self._load_query(asset.value.lower())
        result = self._engine.execute(query)

        # convert row proxies into dictionaries
        assets = [
            {column: value for column, value in rowproxy.items()}
            for rowproxy in result
        ]

        if asset == MetabaseAsset.DATABASE:
            assets = self._database_specifics(assets)

        logger.info(f"Fetching {asset.name} ({len(assets)} results)")

        # keep interesting fields
        return [
            {key: e.get(key) for key in EXPORTED_FIELDS[asset]} for e in assets
        ]



import datetime

from .time_filter import TimeFilter


def test__time_filter(freezer):

    tf = TimeFilter(day=datetime.date(1992, 5, 13))

    assert tf.to_dict() == {
        "day": datetime.date(1992, 5, 13),
        "hour_min": None,
        "hour_max": None,
    }

    freezer.move_to("1989-02-02")

    assert TimeFilter.default().to_dict() == {
        "day": datetime.date(1989, 2, 1),
        "hour_min": 0,
        "hour_max": 23,
    }



"""https://www.notion.so/castordoc/Workshop-Mocking-405ef02712e6446193720abf8d4c2f53"""

from unittest.mock import Mock, patch

from .client import RedshiftClient
from .query import ExtractionQuery

RESULT = [{"foo": "bar"}]
CASTOR_EXTRACTOR_PATH = "source.packages.extractor.castor_extractor"


def fake_execute(statement: str, params: dict):
    return RESULT


@patch.object(RedshiftClient, "_build_uri")
@patch.object(RedshiftClient, "_engine_options")
@patch(f"{CASTOR_EXTRACTOR_PATH}.warehouse.abstract.client.create_engine")
@patch.object(RedshiftClient, "close")
@patch.object(RedshiftClient, "connect")
def test_redshift_client_execute(
    mock_connect,
    mock_close,
    mock_create_engine,
    mock_engine_options,
    mock_build_uri,
):

    mock_connect.return_value = Mock(execute=fake_execute)

    credentials = {}
    client = RedshiftClient(credentials)
    query = ExtractionQuery("toto", {"option": 1})

    res = client.execute(query)

    assert list(res) == RESULT

    mock_close.assert_called_once()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['castor_extractor',
 'castor_extractor.commands',
 'castor_extractor.file_checker',
 'castor_extractor.file_checker.templates',
 'castor_extractor.uploader',
 'castor_extractor.utils',
 'castor_extractor.visualization',
 'castor_extractor.visualization.looker',
 'castor_extractor.visualization.looker.api',
 'castor_extractor.visualization.metabase',
 'castor_extractor.visualization.metabase.client',
 'castor_extractor.visualization.metabase.client.api',
 'castor_extractor.visualization.metabase.client.db',
 'castor_extractor.visualization.mode',
 'castor_extractor.visualization.mode.client',
 'castor_extractor.visualization.qlik',
 'castor_extractor.visualization.qlik.client',
 'castor_extractor.visualization.qlik.client.engine',
 'castor_extractor.visualization.tableau',
 'castor_extractor.visualization.tableau.client',
 'castor_extractor.visualization.tableau.tests',
 'castor_extractor.visualization.tableau.tests.unit',
 'castor_extractor.visualization.tableau.tests.unit.graphql',
 'castor_extractor.visualization.tableau.tests.unit.rest_api',
 'castor_extractor.visualization.tableau.tests.unit.utils',
 'castor_extractor.warehouse',
 'castor_extractor.warehouse.abstract',
 'castor_extractor.warehouse.bigquery',
 'castor_extractor.warehouse.postgres',
 'castor_extractor.warehouse.redshift',
 'castor_extractor.warehouse.snowflake',
 'castor_extractor.warehouse.synapse']

package_data = \
{'': ['*'],
 'castor_extractor.visualization.metabase.client.db': ['queries/*'],
 'castor_extractor.visualization.tableau.tests.unit': ['assets/graphql/metadata/*',
                                                       'assets/rest_api/*'],
 'castor_extractor.warehouse.bigquery': ['queries/*', 'queries/cte/*'],
 'castor_extractor.warehouse.postgres': ['queries/*'],
 'castor_extractor.warehouse.redshift': ['queries/*'],
 'castor_extractor.warehouse.snowflake': ['queries/*'],
 'castor_extractor.warehouse.synapse': ['queries/*']}

install_requires = \
['cachetools>=4.2.4,<5.0.0',
 'certifi==2021.10.8',
 'charset-normalizer>=2.0.7,<3.0.0',
 'google-api-core>=2.1.1,<3.0.0',
 'google-auth>=1.6.3,<3.0.0',
 'google-cloud-bigquery-storage>=2.0.0,<3.0.0',
 'google-cloud-bigquery>=2.0.0,<=3.0.0',
 'google-cloud-core>=2.1.0,<3.0.0',
 'google-cloud-storage>=1.42.3,<2.0.0',
 'google-crc32c>=1.3.0,<2.0.0',
 'google-resumable-media>=2.0.3',
 'googleapis-common-protos>=1.53.0,<7.0.0',
 'idna>=2.5,<2.9',
 'protobuf>=2.0.0,<4.0.0',
 'pyarrow>=6.0.0,<7.0.0',
 'pyasn1-modules>=0.2.8,<1.0.0',
 'pyasn1>=0.4.8,<1.0.0',
 'pycryptodome>=3.0.0,<4.0.0',
 'python-dateutil>=2.0.0,<=3.0.0',
 'pytz==2021.3',
 'requests>=2.0.0,<3.0.0',
 'rsa>=4.0.0,<5.0.0',
 'six>=1.16.0,<2.0.0',
 'sqlalchemy>=1.3,<2.0',
 'tqdm>=4.0.0,<5.0.0',
 'typing-extensions>=3.7.4.3,<4.0.0',
 'urllib3>=1.0.0,<2.0.0']

extras_require = \
{'all': ['looker-sdk>=21.6.0,<=23.0.0',
         'psycopg2-binary>=2.0.0,<3.0.0',
         'snowflake-connector-python>=2.6.2,!=2.7.10',
         'snowflake-sqlalchemy!=1.2.5,<2.0.0',
         'sqlalchemy-bigquery>=1.0.0,<=2.0.0',
         'sqlalchemy-redshift==0.8.2',
         'tableauserverclient==0.17.0',
         'websocket-client<0.55.0'],
 'bigquery': ['sqlalchemy-bigquery>=1.0.0,<=2.0.0'],
 'looker': ['looker-sdk>=21.6.0,<=23.0.0'],
 'metabase': ['psycopg2-binary>=2.0.0,<3.0.0'],
 'qlik': ['websocket-client<0.55.0'],
 'redshift': ['psycopg2-binary>=2.0.0,<3.0.0', 'sqlalchemy-redshift==0.8.2'],
 'snowflake': ['snowflake-connector-python>=2.6.2,!=2.7.10',
               'snowflake-sqlalchemy!=1.2.5,<2.0.0'],
 'tableau': ['tableauserverclient==0.17.0']}

entry_points = \
{'console_scripts': ['castor-extract-bigquery = '
                     'castor_extractor.commands.extract_bigquery:main',
                     'castor-extract-looker = '
                     'castor_extractor.commands.extract_looker:main',
                     'castor-extract-metabase-api = '
                     'castor_extractor.commands.extract_metabase_api:main',
                     'castor-extract-metabase-db = '
                     'castor_extractor.commands.extract_metabase_db:main',
                     'castor-extract-mode = '
                     'castor_extractor.commands.extract_mode:main',
                     'castor-extract-postgres = '
                     'castor_extractor.commands.extract_postgres:main',
                     'castor-extract-qlik = '
                     'castor_extractor.commands.extract_qlik:main',
                     'castor-extract-redshift = '
                     'castor_extractor.commands.extract_redshift:main',
                     'castor-extract-snowflake = '
                     'castor_extractor.commands.extract_snowflake:main',
                     'castor-extract-tableau = '
                     'castor_extractor.commands.extract_tableau:main',
                     'castor-file-check = '
                     'castor_extractor.commands.file_check:main',
                     'castor-upload = castor_extractor.commands.upload:main']}

setup_kwargs = {
    'name': 'castor-extractor',
    'version': '0.0.43',
    'description': 'Extract your metadata assets.',
    'long_description': '# Castor Extractor <img src="https://app.castordoc.com/images/castor_icon_dark.svg" width=30 />\n\n\nThis library contains utilities to extract your metadata assets into `JSON` or `CSV` files, on your local machine.\nAfter extraction, those files can be pushed to Castor for ingestion.\n\n- Visualization assets are typically:\n  - `dashboards`\n  - `users`\n  - `folders`\n  - ...\n\n\n- Warehouse assets are typically:\n  - `databases`\n  - `schemas`\n  - `tables`\n  - `columns`\n  - `queries`\n  - ...\n\nIt also embeds utilities to help you push your metadata to Castor:\n- `File Checker` to validate your [generic](https://docs.castordoc.com/integrations/data-warehouses/generic-warehouse) CSV files before pushing to Castor\n- `Uploader` to push extracted files to our Google-Cloud-Storage (GCS)\n\n\n# Table of contents\n\n\n- [Installation](#installation)\n  * [Create castor-env](#create-castor-env)\n  * [PIP Install](#pip-install)\n  * [Create the output directory](#create-the-output-directory)\n- Documentation\n  - [Extract from BI Tools](https://docs.castordoc.com/castor-package/read-me/bi-tools)\n  - [Extract from Warehouse](https://docs.castordoc.com/castor-package/read-me/warehouse)\n  - [Uploader](https://docs.castordoc.com/castor-package/read-me/uploader)\n  - [File checker](https://docs.castordoc.com/castor-package/read-me/file-checker)\n- [Contact us](#contact-us)\n\n# Installation\n\nRequirements: **python3.7+**\n<img src="https://upload.wikimedia.org/wikipedia/commons/c/c3/Python-logo-notext.svg" width=20 />\n\n\n## Create castor-env\n\nWe advise to create a dedicated [Python environment](https://docs.python.org/3/library/venv.html).\n\nHere\'s an example using `Pyenv` and Python `3.7.9`:\n\n- Install Pyenv\n```bash\nbrew install pyenv\nbrew install pyenv-virtualenv\n```\n\n- [optional] Update your `.bashrc` if you encounter this [issue](https://stackoverflow.com/questions/45577194/failed-to-activate-virtualenv-with-pyenv/45578839)\n```bash\neval "$(pyenv init -)"\neval "$(pyenv init --path)"\neval "$(pyenv virtualenv-init -)"\n```\n\n- [optional] Install python 3.7+\n\n```bash\npyenv versions # check your local python installations\n\npyenv install -v 3.7.9 # if none of the installed versions satisfy requirements 3.7+\n```\n\n- Create your virtual env\n```bash\npyenv virtualenv 3.7.9 castor-env # create a dedicated env\npyenv shell castor-env # activate the environment\n\n# optional checks\npython --version # should be `3.7.9`\npyenv version # should be `castor-env`\n```\n\n## PIP install\n\n⚠️ `castor-env` must be created AND activated first.\n```bash\npyenv shell castor-env\n(castor-env) $ # this means the environment is now active\n```\n\nℹ️ please upgrade `PIP` before installing Castor.\n```\npip install --upgrade pip\n```\n\nRun the following command to install `castor-extractor`:\n```\npip install castor-extractor\n```\n\nDepending on your use-case, you can also install one of the following `extras`:\n```\npip install castor-extractor[looker]\npip install castor-extractor[tableau]\npip install castor-extractor[metabase]\npip install castor-extractor[qlik]\npip install castor-extractor[bigquery]\npip install castor-extractor[redshift]\npip install castor-extractor[snowflake]\n```\n\n## Create the output directory\n\n```bash\nmkdir /tmp/castor\n```\n\nYou will provide this path in `extraction` scripts as following:\n```\ncastor-extract-bigquery --output=/tmp/castor\n```\n\nAlternatively, you can also set the following `ENV` in your `bashrc`:\n```bash\nexport CASTOR_OUTPUT_DIRECTORY="/tmp/castor"\n````\n\n\n# Contact\n\nFor any questions or bug report, contact us at [support@castordoc.com](mailto:support@castordoc.com)\n\n[Castor](https://castordoc.com) helps you find, understand, use your data assets\n',
    'author': 'Castor',
    'author_email': 'support@castordoc.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.castordoc.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import pytest

from .constants import ACCESS_DENIED_ERROR_CODE, APP_SIZE_EXCEEDED_ERROR_CODE
from .error import (
    AccessDeniedError,
    AppSizeExceededError,
    JsonRpcError,
    raise_for_error,
)


def _error(error_code: int) -> dict:
    return {
        "message": "Houston, we have a problem",
        "code": error_code,
    }


def test_raise_for_error():
    any_message = {"hello": "world"}

    # no error
    response = {"everything is": "OK"}
    raise_for_error(any_message, response)

    # error
    response = {"error": _error(12)}
    with pytest.raises(JsonRpcError):
        raise_for_error(any_message, response)

    # access denied error
    response = {"error": _error(ACCESS_DENIED_ERROR_CODE)}
    with pytest.raises(AccessDeniedError):
        raise_for_error(any_message, response)

    # source size exceeded error
    response = {"error": _error(APP_SIZE_EXCEEDED_ERROR_CODE)}
    with pytest.raises(AppSizeExceededError):
        raise_for_error(any_message, response)



import logging
from typing import Callable, List, Optional, Tuple, Type, Union

logger = logging.getLogger(__name__)


class SafeMode:
    """
    SafeMode class to parameterize safe_mode decorator

    Attributes:
        self.exceptions: tuple of exception that will be caught
        self.max_errors: nb of errors to catch
        self.errors_caught : list of errors caught
    """

    def __init__(
        self,
        exceptions: Tuple[Type[BaseException], ...],
        max_errors: Union[int, float],
    ):
        self.exceptions = exceptions
        self.max_errors = max_errors
        self.errors_caught: List[Type[BaseException]] = []

    @property
    def should_raise(self) -> bool:
        return len(self.errors_caught) >= self.max_errors


def safe_mode(
    safe_mode_params: Optional[SafeMode] = None,
    default: Optional[Callable] = None,
):
    """
    safe_mode decorator

    Args : safe_mode_params: param exception and number of exceptions to
    catch default: Optional - default function to execute if decorated
    function raises exception

    Note: KeyboardInterrupt is excluded from being excepted by design.
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            if safe_mode_params is None:
                return function(*args, **kwargs)

            try:
                result = function(*args, **kwargs)
            except KeyboardInterrupt:
                raise
            except safe_mode_params.exceptions as e:
                if safe_mode_params.should_raise:
                    raise e
                logger.error(
                    f"Safe mode : skip error {e} in function {function.__name__} with args {args} kwargs {kwargs}"
                )
                logger.debug(e, exc_info=True)
                safe_mode_params.errors_caught.append(e)
                return default() if default else None
            return result

        return wrapper

    return decorator



from typing import List, Optional

from ..abstract import (
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
    WarehouseAsset,
)


class RedshiftQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from Redshift.
    """

    def __init__(
        self,
        time_filter: Optional[TimeFilter] = None,
    ):
        super().__init__(time_filter=time_filter)

    def build(self, asset: WarehouseAsset) -> List[ExtractionQuery]:
        query = self.build_default(asset)
        return [query]



"""https://www.notion.so/castordoc/Workshop-Mocking-405ef02712e6446193720abf8d4c2f53"""

from typing import List, NamedTuple, Set, Tuple
from unittest.mock import Mock, patch

from source.packages.extractor import BigQueryClient
from source.packages.extractor.castor_extractor.warehouse.bigquery import (
    extract_all,
)


class MockProject(NamedTuple):
    project_id: str


class MockDataset(NamedTuple):
    dataset_id: str


client_project_structure = {
    "project_1": [MockDataset("dataset_1"), MockDataset("dataset_2")],
    "project_2": [MockDataset("dataset_3"), MockDataset("hidden_dataset")],
    "hidden_project": [MockDataset("dataset_h1"), MockDataset("dataset_h2")],
}


class MockBigQueryClient(BigQueryClient):
    def __init__(self):
        self._db_allowed = ["project_2", "project_1"]
        self._dataset_blocked = ["hidden_dataset"]
        self._db_blocked = ["hidden_project"]

    def _google_cloud_client(self) -> Mock:
        fake_client = Mock()
        fake_client.list_projects = Mock(
            return_value=[
                MockProject("project_1"),
                MockProject("project_2"),
                MockProject("hidden_project"),
            ]
        )
        fake_client.list_datasets = lambda project: client_project_structure[
            project
        ]
        return fake_client


def test_bigquery_client__list_datasets():
    client = MockBigQueryClient()
    assert client._list_datasets() == [
        MockDataset("dataset_1"),
        MockDataset("dataset_2"),
        MockDataset("dataset_3"),
    ]
